#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>

#define UP	1
#define DOWN	2
#define LEFT	3
#define RIGHT	4

#define TAB     8

WINDOW * mainwin;

/* move the cursor in one of four directions */
void mv(int direction, int dist) {

  int x, y, h, w;
  getyx( mainwin, y, x);
  getmaxyx( mainwin, h, w);

  switch (direction) {
    case UP:
      y = y - dist; 
      break;
    case DOWN:
      y = y + dist; 
      break;
    case LEFT:
      x = x - dist; 
      break;
    case RIGHT:
      x = x + dist; 
      break;
  }

  /* keep cursor in bounds */
  if (x >= w-2) x = w - 2;
  if (y >= h-2) y = h - 2;
  if (x < 1) x = 1;
  if (y < 1) y = 1;
  //later these should scroll it

  move(y,x);

}

/* fills a rect with a char */
void fill(char key, int x, int y, int w, int h) {
  int i, _x, _y;
  char str[w+20];

  getyx( mainwin, _y, _x ); //so we can reset the cursor after drawing
  
  for (i = 0; i <= w ; i++) str[i] = key; 
  str[w+1] = '\0';

  for (i = 0; i <= h ; i++) {
    move(y+i, x);
    addstr(str);
  }

  move(_y,_x); //bring cursor back to where it was when we started filling
}

/* start or end marking a region */
void mark(bool *marking, int *lastX, int *lastY) {
  int x, y, startX, startY, width, height;
  getyx( mainwin, y, x );

  if (*marking) {    /* fill region */

    //get dimensions for fill()
    if (*lastX < x) {
      startX = *lastX;
      width = x - *lastX;
    } else {
      startX = x;
      width = *lastX - x;
    }

    if (*lastY < y) {
      startY = *lastY;
      height = y - *lastY; 
    } else {
      startY = y;
      height = *lastY - y;
    }
    
    /* ask for input */
    char k = getch();
    //printf("x: %d\ny: %d\nw: %d\n%h: %d\n", startX, startY, width, height);
    fill(k, startX, startY, width, height);
    
    //reset marking
    *marking = false;

    refresh();
  } else {
    //save and mark current space
    *lastX = x;
    *lastY = y;

    *marking = true;
    attron(A_REVERSE); //blink moves cursors, sometimes doubledraws *
    char k = inch();
    delch();
    insch(k);
    attroff(A_REVERSE);
  }
}

/* goes to first or last space on a line/col */
void homeend(dir) {
  int x,y,h,w;
  getyx( mainwin, y, x );
  getmaxyx( mainwin, h, w);

  switch( dir ) {
    case UP:
      y = 1;
      break;
    case DOWN:
      y = h-2;
      break;
    case LEFT:
      x = 1;
      break;
    case RIGHT:
      x = w-2;
      break;
  }

  move(y, x);
}

void save() {
  FILE *f;
  f = fopen("saved_dungeon.txt", "w+b");
    putwin(stdscr, f);
  //scr_dump("saved_dungeon.txt");
  fclose(f);
}

//moves cursor in direction till it lands on a new char
void chtab(dir) {
  char k = inch();
  char _k = inch();
  int x, y, h, w;
  getyx( mainwin, y, x );
  getmaxyx( mainwin, h, w);


  while ( k==_k && x>0 && y>0 && x<w-1 && y<h-1 ) {
    move(y, x);
    switch( dir ) {
      case UP:
	y--;
	break;
      case DOWN:
	y++;
	break;
      case LEFT:
	x--;
	break;
      case RIGHT:
	x++;
	break;
    }
    _k = inch();
  }
}

int main() {

  /* init ncurses */
  if ( (mainwin = initscr()) == NULL ) {
    perror("cannot init ncurses");
    exit(EXIT_FAILURE);
  }

  clear();
  noecho();
  cbreak();
  keypad(stdscr, TRUE);
  box(mainwin, ACS_VLINE, ACS_HLINE);
  move(1,1);

  /* my vars and flags */
  int looping = 1;
  bool marking = false;
  int lastX = 0;
  int lastY = 0;

  /* loop for input */
  while ( looping ) {
    switch ( getch() ) {

      case KEY_UP:
      case 'i':
	mv(UP, 1);
	break;

      case 'I':
	mv(UP, TAB);
	break;

      case KEY_DOWN:
      case 'k':
	mv(DOWN, 1);
	break;

      case 'K':
	mv(DOWN, TAB);
	break;

      case KEY_LEFT:
      case 'j':
	mv(LEFT, 1);
	break;

      case 'J':
	mv(LEFT, TAB);
	break;

      case KEY_RIGHT:
      case 'l':
	mv(RIGHT, 1);
	break;

      case 'L':
	mv(RIGHT, TAB);
	break;

      case 'o':
	chtab(RIGHT);
	break;

      case 'u':
	chtab(LEFT);
	break;

      case 'y':
	chtab(UP);
	break;

      case 'h':
	chtab(DOWN);
	break;

      case 'O':
	homeend(RIGHT);
	break;

      case 'U':
	homeend(LEFT);
	break;

      case 'Y':
	homeend(UP);
	break;

      case 'H':
	homeend(DOWN);
	break;

      case ' ':
	mark(&marking, &lastX, &lastY);
	break;

      case 'Q':
      case 'q':
	looping = 0;
	break;

      case 's':
	save();
	break;
    }

    refresh();
  }

  /* shut it down */
  delwin(mainwin);
  endwin();
  refresh();

  return EXIT_SUCCESS;
}
